#!/usr/bin/env python3
"""
A Platform that uses SLURM representation.

This file contains a python Object-Oriented representation of a Platform that uses SLURM as its
scheduler.

Two supporting classes are also defined here:
    SlurmConf
    Node

NOTE: We only support parsin slurm.conf files for slurm versions >= 19.05

BSC-CNS ES
Manuel Giménez de Castro Marciani
2022
"""

# this module is just to split key names in the slurm.conf file in the capitalized letters
import re


def slurm_strtimetint(slurm_time: str) -> int:
    """
    str time to int (strtimetint) following datetime conventions
    This function converts slurm's time format for Decay, MaxAgeTime to an integer representing seconds
    :param slurm_time: string with the param time
    :return: integer with the converted time
    """
    days, hours, minutes = 0, 0, 0  # initialize every meaningful parameter to 0
    # we have days-hours:minutes
    if "-" in slurm_time and ":" in slurm_time:
        days, tail = slurm_time.split("-")
        hours, minutes, *tail = tail.split(":")
    # we have days-hours
    elif "-" in slurm_time and ":" not in slurm_time:
        days, hours = slurm_time.split("-")
    # we have HOURS:MINUTES:
    elif "-" not in slurm_time and ":" in slurm_time:
        hours, minutes, *tail = slurm_time.split(":")
    # if there is no - or : then we have minutes ONLY
    else:
        minutes = int(slurm_time)
    # WE NEVER HAVE SECONDS
    # days multiplied by the number of seconds in a day, hours multiplied by the number of seconds in 1 hour, etc.
    # if the cast failed is because we have an incorrect time format
    return int(days)*86400+int(hours)*3600+int(minutes)*60


class Node:
    """
    This class represents a single node as specified on slurm.conf
    """
    def __init__(self, name: str, sockets: int, cores_per_socket: int, mem_spec_limit: int,
                 threads_per_core: int, real_memory: int, weight: int, gres=None, features: str = ''):
        if gres is None:
            gres = {}
        self._name = name
        self._sockets = sockets
        self._coresPerSocket = cores_per_socket
        self._threadsPerCore = threads_per_core
        self._memSpecLimit = mem_spec_limit  # memory reserved for SLURM in MiB
        self._realMemory = real_memory  # amount of available memory in nodes
        self._gres = gres  # dictionary with generic resources (gres)
        self._weight = weight
        self._features = features  # arbitrary comments

    def __str__(self):
        output = f"Node {self.name}  Number of Sockets: {self.sockets}  Cores per Socket: {self.cores_per_socket}  " \
                 f"Threads per Core: {self.threads_per_core}  Slurm reserved memory: {self.mem_spec_limit}  " \
                 f"Real memory: {self.real_memory}"
        if len(self.gres) != 0:
            output += f"  Generic Resources: {[res for res in self.gres]}"
        if self.weight != 0:
            output += f"  Weight: {self.weight}"
        if self.features != '':
            output += f"  Features:{self.features}"
        return output

    @property
    def name(self) -> str:
        """

        :return: Name of the Node
        """
        return self._name

    @property
    def sockets(self) -> int:
        """

        :return: Number of sockets in this Node
        """
        return self._sockets

    @property
    def cores_per_socket(self) -> int:
        """

        :return: Number of cores per socket in this Node
        """
        return self._coresPerSocket

    @property
    def threads_per_core(self) -> int:
        """

        :return: Number of threads per Core in this Node
        """
        return self._threadsPerCore

    @property
    def mem_spec_limit(self) -> int:
        """
        Returns memory allocated in each Node for slurm.
        :return: Reserved Memory in Megabytes for slurm
        """
        return self._memSpecLimit

    @property
    def real_memory(self) -> int:
        """

        :return: Total main memory (RAM) in the node
        """
        return self._realMemory

    @property
    def gres(self) -> dict:
        """
        Generic Resources (GRES) are custom resources that slurm administrators can configure slurm to consider in
        scheduling and accounting. Usually they indicate GPUs, but it could be any other type of accelerators, storage,
        network devices etc. In our case we will store them in a dictionary with all associated parameters as values.
        :return: dictionary with generic resources
        """
        return self._gres

    @property
    def features(self) -> str:
        """
        Features are analogous to comments in programming. Slurm administrator can configure for each Node as they wish.
        For example, in LUMI they indicate the processor in the particular Node. In MN4 they are not present.
        :return: Features string
        """
        return self._features

    @property
    def weight(self):
        """

        :return: Weight of using this specific node in the priority computation
        """
        return self._weight


class SlurmConf:
    """
    This class handles slurm.conf files. Since we are only interested in scheduling and priority, many of its
    configurations will be pruned.

    # attribute names follow the same name as in the slurm conf!

    WE ONLY SUPPORT SLURM VERSION >= 19.05 because since then the default fair share algorithm is FAIR TREE.
    """

    def __init__(self):
        self._priorityType = ''
        self._priorityFlags = ''
        self._priorityFavorSmall = False  # If we flip the computation for job size factor
        self._priorityUsageResetPeriod = ''  # interval between USAGES reset
        self._calcPeriod = 0  # Period to compute decay of USAGE
        self._priorityDecayHalfLife = 0  # decay factor to compute USAGE
        self._priorityMaxAge = 0  # type it takes for a job to max out its AGE factor
        # WEIGHTS
        # TODO might need to add more, this are the only nonzero in MN4!
        self._priorityWeightFairShare = 0
        self._priorityWeightAge = 0
        self._priorityWeightQOS = 0
        self._priorityWeightJobSize = 0
        self._priorityWeightPartition = 0
        # Compute NODES
        self._nodes = []  # array with nodes
        # TODO How much info do we keep from slurm conf????????????

    def parse_from_file(self, slurm_conf_path):
        """
        This routine parses a real slurm.conf file a creates an object type SlurmConf with the
        meaningful parameters for priority and scheduling
        :param slurm_conf_path: location of the slurm.conf file
        :return: None
        """
        with open(slurm_conf_path, "r") as slurmconf:
            # initialize every Default NODE attribute to 0
            default_sockets = 0
            default_cores_per_socket = 0
            default_threads_per_core = 0
            default_real_memory = 0
            default_mem_spec_limit = 0
            default_weight = 0
            for slurmline in slurmconf:
                # avoid comment lines
                if slurmline[0] != "#":
                    # if the line is one with definitions
                    if "=" in slurmline:
                        # split at the first appearance of equal symbol
                        key, value = slurmline.split("=", 1)
                        # we make case-insensitive comparisons with the fields we defined in the SlurmConf class
                        slurm_conf_attributes = [field.strip("_").lower() for field in self.__dict__.keys()]
                        if key.lower() in slurm_conf_attributes:
                            # transforms from PriorityFlag to priority_flag
                            # manuel: HARDCODED FOR PriorityWeightQOS, cuz otherwise it would add a
                            # new attribute priority_weight_q_o_s... ugly code
                            if key == "PriorityWeightQOS":
                                self.priority_weight_qos = int(value.strip("\n"))
                            else:
                                field_key = '_'.join(re.findall('[A-Z][^A-Z]*', key)).lower()
                                # remove the new line from variable value
                                # cast and set the new value to field_key
                                self.cast_set_attribute(field_key, value.strip("\n"))
                        # manuel: Our implementation doesn't contain a NodeName attribute, thus we need
                        # to add manually... ugly code.
                        elif key == "NodeName":
                            # if this is a default node, we keep parameters for the subsequent nodes in the list
                            # keep in mind that this only make sense if we are reading from top to bottom the slurm.conf
                            # any other order will break this code
                            node_values = slurmline.split(" ")  # split into every attribute of the node
                            if "DEFAULT" in value:
                                for node_field in node_values:
                                    node_key, node_value = node_field.split("=")
                                    if node_key.lower() == "sockets":
                                        default_sockets = int(node_value)
                                    elif node_key.lower() == "corespersocket":
                                        default_cores_per_socket = int(node_value)
                                    elif node_key.lower() == "threadspercore":
                                        default_threads_per_core = int(node_value)
                                    elif node_key.lower() == "realmemory":
                                        default_real_memory = int(node_value)
                                    elif node_key.lower() == "weight":
                                        default_weight = int(node_value)
                                    elif node_key.lower() == "memspeclimit":
                                        default_mem_spec_limit = int(node_value)
                            else:
                                # per node attributes
                                sockets = 0
                                cores_per_socket = 0
                                threads_per_core = 0
                                real_memory = 0
                                mem_spec_limit = 0
                                weight = 0
                                features = ''
                                gres = ''
                                node_name_list = []  # list which contains all the names of the nodes
                                for node_field in node_values:
                                    node_key, node_value = node_field.split("=")
                                    if node_key.lower() == "sockets":
                                        sockets = int(node_value)
                                    elif node_key.lower() == "corespersocket":
                                        cores_per_socket = int(node_value)
                                    elif node_key.lower() == "threadspercore":
                                        threads_per_core = int(node_value)
                                    elif node_key.lower() == "realmemory":
                                        real_memory = int(node_value)
                                    elif node_key.lower() == "weight":
                                        weight = int(node_value)
                                    elif node_key.lower() == "features":
                                        features = node_value
                                    elif node_key.lower() == "gres":
                                        gres = node_value
                                    # in slurm is possible to define multiple nodes in one line by issuing
                                    # prefix[0-10] which will create prefix0, prefix1, prefix2, ..., prefix10
                                    elif node_key.lower() == "nodename":
                                        if "[" in node_value:
                                            prefix, rest = node_value.split("[")  # the suffix will be to the right of [
                                            # this whole part is to remove the ']' and move through every min-max range
                                            rest = rest.strip("]")
                                            for ranges in rest.split(","):
                                                # casts to int and unpack to bottom and top
                                                bottom, top = map(int, ranges.split("-"))
                                                if bottom >= top:
                                                    raise ValueError(f"Node definition is wrong. "
                                                                     f"Invalid interval found in {prefix}, "
                                                                     f"{bottom}-{top}")
                                                node_name_list += [f"{prefix}{val}" for val in range(bottom, top+1)]
                                        else:
                                            # we might have just a simple node name
                                            node_name_list.append(node_value)

                                # if we can't find these values in the node name definition, then it is inherited
                                # from the default node
                                if sockets == 0:
                                    sockets = default_sockets
                                if cores_per_socket == 0:
                                    cores_per_socket = default_cores_per_socket
                                if threads_per_core == 0:
                                    threads_per_core = default_threads_per_core
                                if mem_spec_limit == 0:
                                    mem_spec_limit = default_mem_spec_limit
                                if real_memory == 0:
                                    real_memory = default_real_memory
                                if weight == 0:
                                    weight = default_weight
                                # go through every node name creating a node
                                for node_name in node_name_list:
                                    self.add_node(Node(node_name, sockets, cores_per_socket, mem_spec_limit,
                                                       threads_per_core, real_memory, weight, gres, features))

    def cast_set_attribute(self, attribute_name: str, value: str):
        """
        This function transforms the value (string) to the correct value for the field as specified by the
        setter function. Then we add this attribute
        :param attribute_name: SlurmConf attribute name that will receive the value
        :param value: value to set in attribute_name
        :return: None
        """
        # we get the function and check its type hints
        setter_func = getattr(self, attribute_name)
        # iterate over all type hints (e.g. the type after ':' in the function declaration)
        for param_name in setter_func.__annotations__:
            # if it is not the correct type, we will have to transform/cast it
            if type(value) != setter_func.__annotations__[param_name]:
                if param_name == "period":
                    value = slurm_strtimetint(value)  # transform string to seconds
                # case this is the favor small option
                elif param_name == "option":
                    if value.lower() == "yes":
                        value = True
                    else:
                        value = False
                elif param_name == "weight":
                    value = int(value)  # simple cast
                # we don't handle node input here
                # finally, after casting, we call the setter function
                setattr(self, attribute_name, value)

    @property
    def priority_type(self) -> str:
        """
        Outputs the priority type (e.g. multifactor)
        :return: Priority type
        """
        return self._priorityType

    @priority_type.setter
    def priority_type(self, priority: str) -> None:
        """
        Sets priority type of the scheduler
        :param priority: type of priority (e.g. multifactor)
        :return: None
        """
        self._priorityType = priority

    @property
    def priority_flags(self) -> str:
        """
        Outputs flags in the priority (e.g. NO_FAIR_TREE to select Classic Fairshare instead of Fairtree)
        :return: Priority Flags
        """
        return self._priorityFlags

    @priority_flags.setter
    def priority_flags(self, flags: str) -> None:
        """
        Sets priority flags
        :param flags: priority flags (e.g. NO_FAIR_TREE)
        :return: None
        """
        self._priorityFlags = flags

    @property
    def priority_favor_small(self) -> bool:
        """
        Returns if the scheduler favors small (that is, small jobs get factor 1 whilst big jobs are penalized)
        :return: True if scheduler favor small jobs
        """
        return self._priorityFavorSmall

    @priority_favor_small.setter
    def priority_favor_small(self, option: bool):
        """
        Sets if the scheduler favors small (that is, small jobs get factor 1 whilst big jobs are penalized)
        :param option: boolean that says whether the scheduler favors small jobs
        :return: None
        """
        self._priorityFavorSmall = option

    @property
    def priority_usage_reset_period(self) -> str:
        """
        Returns the interval in which the usage is reset. Can only be
        "NONE": Never resets;
        "NOW": Resets every startup;
        "DAILY": Resets every day at midnight;
        "WEEKLY": Resets every week on Sunday at 00:00;
        "MONTHLY":  Resets every 1st day at 00:00;
        "QUARTERLY": Resets on the first day of each quarter at 00:00;
         "YEARLY": Resets on the first day of each year at 00:00.
        :return: interval in which usage is reset
        """
        return self._priorityUsageResetPeriod

    @priority_usage_reset_period.setter
    def priority_usage_reset_period(self, reset_period: str) -> None:
        """
        Sets the interval in which usage is reset. Can only be
        "NONE": Never resets;
        "NOW": Resets every startup;
        "DAILY": Resets every day at midnight;
        "WEEKLY": Resets every week on Sunday at 00:00;
        "MONTHLY":  Resets every 1st day at 00:00;
        "QUARTERLY": Resets on the first day of each quarter at 00:00;
         "YEARLY": Resets on the first day of each year at 00:00.
        :param reset_period: interval in which usage is reset
        :return: None
        """
        if reset_period not in ["NONE", "NOW", "DAILY", "WEEKLY", "MONTHLY", "QUARTERLY", "YEARLY"]:
            raise ValueError(f"{reset_period} not a valid option.")
        self._priorityUsageResetPeriod = reset_period

    @property
    def calc_period(self) -> int:
        """
        :return: Time in minutes in which fair tree algorithm and decay is applied to usage
        """
        return self._calcPeriod

    @calc_period.setter
    def calc_period(self, period: int) -> None:
        """
        Sets the period in seconds which fair tree algorithm and decay is computed.
        :param period: Time in minutes in which decay is applied
        :return: None
        """
        if period < 0:
            raise ValueError(f"{period} not a valid option.")
        self._calcPeriod = period

    @property
    def priority_decay_half_life(self) -> int:
        """
        Returns the half life period in seconds of the usage. This is implemented to account more for recent usage, as
        opposed to give the same weight for the whole usage period.
        :return: Half lifetime in seconds of usage
        """
        return self._priorityDecayHalfLife

    @priority_decay_half_life.setter
    def priority_decay_half_life(self, period: int) -> None:
        """
        Sets half life period in minutes of the usage.
        :param period: Period in minutes in which decay is computed in priority
        :return: None
        """
        if period < 0:
            raise ValueError(f"{period} not a valid option.")
        self._priorityDecayHalfLife = period

    @property
    def priority_max_age(self) -> int:
        """
        :return: Time in seconds to reach max age factor
        """
        return self._priorityMaxAge

    @priority_max_age.setter
    def priority_max_age(self, period: int) -> None:
        """
        Sets time in seconds in which age factor maximizes.
        :param period: Time in seconds to reach max age factor
        :return: None
        """
        if period < 0:
            raise ValueError(f"{period} not a valid option.")
        self._priorityMaxAge = period

    @property
    def priority_weight_fairshare(self) -> int:
        """
        :return: Weight in priority of fair share factor
        """
        return self._priorityWeightFairShare

    @priority_weight_fairshare.setter
    def priority_weight_fairshare(self, weight: int) -> None:
        """
        Sets fair share weight in the priority computation.
        :param weight: Weight of fair share in priority
        :return: None
        """
        if type(weight) is not int:
            print("Casting fair share weight to integer")
            weight = int(weight)
        if weight < 0:
            raise ValueError(f"{weight} not a valid priority weight option.")
        self._priorityWeightFairShare = weight

    @property
    def priority_weight_age(self) -> int:
        """
        :return: Weight of age in priority
        """
        return self._priorityWeightAge

    @priority_weight_age.setter
    def priority_weight_age(self, weight: int) -> None:
        """
        :param weight: Weight of age in priority computation
        :return: None
        """
        if type(weight) is not int:
            print("Casting age weight to integer")
            weight = int(weight)
        if weight < 0:
            raise ValueError(f"{weight} not a valid priority weight option.")
        self._priorityWeightAge = weight

    @property
    def priority_weight_qos(self) -> int:
        """
        :return: Weight of QOS in priority
        """
        return self._priorityWeightQOS

    @priority_weight_qos.setter
    def priority_weight_qos(self, weight: int) -> None:
        """
        Sets QOS weight in the priority computation.
        :param weight: Weight of QOS in priority
        :return: None
        """
        if type(weight) is not int:
            print("Casting qos weight to integer")
            weight = int(weight)
        if weight < 0:
            raise ValueError(f"{weight} not a valid priority weight option.")
        self._priorityWeightQOS = weight

    @property
    def priority_weight_jobsize(self) -> int:
        """
        Returns weight in priority of job size (in this case, job size only considers CPU count).
        :return: Weight of job size in priority
        """
        return self._priorityWeightJobSize

    @priority_weight_jobsize.setter
    def priority_weight_jobsize(self, weight: int) -> None:
        """
        Sets job size weight in the priority computation. Job size in this case only considers CPU count, not
        runtime.
        :param weight: Weight of job size in priority
        :return: None
        """
        if type(weight) is not int:
            print("Casting job size weight to integer")
            weight = int(weight)
        if weight < 0:
            raise ValueError(f"{weight} not a valid priority weight option.")
        self._priorityWeightJobSize = weight

    @property
    def priority_weight_partition(self) -> int:
        """
        :return: Weight of partition in priority calculation
        """
        return self._priorityWeightPartition

    @priority_weight_partition.setter
    def priority_weight_partition(self, weight: int) -> None:
        """
        Sets the weight of partition in the priority calculation.
        :param weight: Weight of partition in the priority calculation
        :return: None
        """
        if type(weight) is not int:
            print("Casting partition weight to integer")
            weight = int(weight)
        if weight < 0:
            raise ValueError(f"{weight} not a valid priority weight option.")
        self._priorityWeightPartition = weight

    @property
    def nodes(self) -> list:
        """
        Returns list with elements type Node, which encodes all info about hardware parameters of the machine that
        slurm is aware of
        :return: list with elements type Node
        """
        return self._nodes

    def add_node(self, new_node: Node) -> None:
        """
        Includes a new node the list of nodes
        :param new_node: new computing Node
        :return: None
        """
        self._nodes.append(new_node)


class Platform:
    """
    This class a Platform: its slurm.conf and meaningful hardware parameters.
    """
    def __init__(self):
        self._slurmConf = SlurmConf()  # empty config file
